from aiographql.client.client import GraphQLClient, GraphQLQueryMethod
from aiographql.client.error import GraphQLError
from aiographql.client.exceptions import (
    GraphQLClientException,
    GraphQLClientValidationException,
    GraphQLRequestException,
    GraphQLIntrospectionException,
)
from aiographql.client.request import GraphQLRequest
from aiographql.client.response import GraphQLResponse
from aiographql.client.subscription import (
    GraphQLSubscription,
    GraphQLSubscriptionEvent,
    GraphQLSubscriptionEventType,
)

__all__ = [
    i.__name__
    for i in [
        GraphQLClient,
        GraphQLQueryMethod,
        GraphQLError,
        GraphQLClientException,
        GraphQLClientValidationException,
        GraphQLRequestException,
        GraphQLIntrospectionException,
        GraphQLRequest,
        GraphQLResponse,
        GraphQLSubscription,
        GraphQLSubscriptionEvent,
        GraphQLSubscriptionEventType,
        GraphQLQueryMethod,
    ]
]



from dataclasses import dataclass, field
from typing import Dict, Any, Optional, List, Union


@dataclass(frozen=True)
class GraphQLError:
    """
    GraphQL error response object.
    """

    extensions: Dict[str, Any] = field(default_factory=dict)
    locations: Optional[List[Dict[str, int]]] = field(default=None)
    message: Optional[str] = field(default=None)
    path: Optional[List[Union[str, int]]] = field(default=None)



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Union

import aiohttp
import graphql
from cafeteria.asyncio.callbacks import CallbackRegistry, CallbackType

from aiographql.client.exceptions import (
    GraphQLClientException,
    GraphQLClientValidationException,
    GraphQLIntrospectionException,
    GraphQLRequestException,
)
from aiographql.client.helpers import create_default_connector
from aiographql.client.request import GraphQLRequest
from aiographql.client.response import GraphQLResponse
from aiographql.client.subscription import (
    CallbacksType,
    GraphQLSubscription,
    GraphQLSubscriptionEventType,
)


@dataclass(frozen=True)
class GraphQLQueryMethod:
    post: str = "post"
    get: str = "get"


class GraphQLClient:
    """
    Client implementation handling all interactions with a specified endpoint. The
    following example shows how to make a simple query.

    .. code-block:: python

        client = GraphQLClient(
            endpoint="http://127.0.0.1:8080/v1/graphql",
            headers={"Authorization": "Bearer <token>"},
        )
        response: GraphQLResponse = await client.query("{ city { name } }")

    You can also use an application scoped :class:`aiohttp.ClientSession` throughout
    the life of the client as show below.

    .. code-block:: python
        :emphasize-lines: 1,4

        async with aiohttp.ClientSession() as session:
            client = GraphQLClient(
                endpoint="http://127.0.0.1:8080/v1/graphql",
                session=session
            )

    :param endpoint: URI of graph api.
    :param headers: Default headers to use for every request made by this client.
        By default the client adds 'Content-Type: application/json' and
        'Accept-Encoding: gzip' to all requests. These can be overridden by
        specifying then here.
    :param method: Default method to use when submitting a GraphQL request to the
        specified `endpoint`.
    :param session: Optional `aiohttp.ClientSession` to use when making requests.
        This is expected to be externally managed.
    """

    def __init__(
        self,
        endpoint: str,
        headers: Optional[Mapping[str, str]] = None,
        method: Optional[str] = None,
        schema: Optional[graphql.GraphQLSchema] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        self.endpoint = endpoint
        self._method = method or GraphQLQueryMethod.post
        self._headers = {"Content-Type": "application/json", "Accept-Encoding": "gzip"}
        self._headers.update(headers or dict())
        self._schema = schema
        self._session = session

    async def introspect(
        self, headers: Optional[Dict[str, str]] = None
    ) -> graphql.GraphQLSchema:
        """
        Introspect the GraphQL endpoint specified for this client and return a `graphql.GraphQLSchema` object
        specifying the schema associated with this endpoint.

        :return: GraphQL schema for the configured endpoint
        """
        request = GraphQLRequest(
            query=graphql.get_introspection_query(descriptions=False),
            validate=False,
            headers=headers,
        )
        introspection = await self.query(request)
        try:
            return graphql.build_client_schema(introspection=introspection.data)
        except TypeError:
            raise GraphQLIntrospectionException(
                f"Failed to build schema from introspection data: {introspection.errors}"
            )

    async def get_schema(
        self, refresh: bool = False, headers: Optional[Dict[str, str]] = None
    ) -> graphql.GraphQLSchema:
        """
        Get the introspected schema for the endpoint used by this client. If an unexpired cache exists, this is
        returned unless the `refresh` parameter is set to True.

        :param refresh: Refresh the cached schema by forcing an introspection of the GraphQL endpoint.
        :param headers: Request headers
        :return: The GraphQL schema as introspected. This maybe a previously cached value.
        """
        # TODO: consider adding ttl logic for expiring schemas for long running services
        if self._schema is None or refresh:
            self._schema = await self.introspect(headers=headers)
        return self._schema

    async def validate(
        self,
        request: GraphQLRequest,
        schema: Optional[graphql.GraphQLSchema] = None,
        headers: Optional[Dict[str, str]] = None,
        force: bool = False,
    ) -> None:
        """
        Validate a given request against a schema (provided or fetched). Validation is
        skipped if the request's `validate` property is set to `False` unless forced.

        :param request: Request that is to be validated.
        :param schema: Schema against which provided request should be validated, if
                       different from `GraphQLRequest.schema` or as fetched from the
                       client endpoint.
        :param headers: Headers to be set when fetching the schema from the client
                        endpoint. If provided, request headers are ignored.
        :param force: Force validation even if the provided request has validation
                      disabled.
        """
        if not request.validate and not force:
            # skip validation if request validate flag is set to false
            return

        schema = schema or await self.get_schema(headers=headers or request.headers)
        errors = await asyncio.get_running_loop().run_in_executor(
            None, graphql.validate, schema, graphql.parse(request.query)
        )
        if errors:
            raise GraphQLClientValidationException(*errors)

    def _prepare_request(
        self,
        request: Union[GraphQLRequest, str],
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> GraphQLRequest:
        """
        Helper method to ensure that queries handle both string and
        :class:`GraphQLRequest` objects.

        :param request: Request to send to the GraphQL server.
        :param operation: GraphQL operation name to use if the `GraphQLRequest.query`
                          contains named operations. This will override any default
                          operation set.
        :param variables: Query variables to set for the provided request. This will
                          override the default values for any existing variables in the
                          request if set.
        :param headers: Additional headers to be set when sending HTTP request.
        :return: A copy of the `request` object with the specified values of
            `operation`, `variables` and `headers` set/merged.
        """
        if isinstance(request, str):
            request = GraphQLRequest(query=request)

        return request.copy(
            headers=headers,
            headers_fallback=self._headers,
            operation=operation,
            variables=variables,
        )

    async def _http_request(
        self,
        session: aiohttp.ClientSession,
        method: str,
        request: GraphQLRequest,
        **kwargs: Any,
    ):
        """
        Helper method to make an http request using the provided *session*.

        :param session: Session to use when making the request.
        :param method: HTTP method to use when making the request.
        :param request: Prepared GraphQL request to dispatch to the server.
        :param kwargs: Additional arguments to pass to
            :method:`aiohttp.ClientSession.request` when making the request.
        :raises: :class:`GraphQLRequestException` when the server responds with a
            non 200 status code.
        :return: Query response.
        """
        async with session.request(
            method=method, url=self.endpoint, headers=request.headers, **kwargs
        ) as resp:
            body = await resp.json()
            response = GraphQLResponse(request=request, json=body)

            if 200 <= resp.status < 300:
                return response

            raise GraphQLRequestException(response)

    async def query(
        self,
        request: Union[GraphQLRequest, str],
        method: str = None,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> GraphQLResponse:
        """
        Method to send provided :class:`GraphQLRequest` to the configured endpoint as
        an HTTP request. This method handles the configuration of headers HTTP method
        specific handling of request parameters and/or data as required.

        The order of precedence, least to greatest, of headers is as follows,
            1. client headers (:attr:`GraphQLClient.headers`)
            2. request headers (:attr:`GraphQLRequest.headers`)
            3. `headers` specified as method parameter

        In accordance to the GraphQL specification, any non 2XX response  is treated as
        an error and raises `GraphQLTransactionException` instance.

        :param request: Request to send to the GraphQL server.
        :param method: HTTP method to use when submitting request (POST/GET). If once is
                       not specified, the client default (`GraphQLClient.method`) is
                       used.
        :param headers: Additional headers to be set when sending HTTP request.
        :param operation: GraphQL operation name to use if the `GraphQLRequest.query`
                          contains named operations. This will override any default
                          operation set.
        :param variables: Query variables to set for the provided request. This will
                          override the default values for any existing variables in the
                          request if set.
        :param session: Optional `aiohttp.ClientSession` to use for requests
        :return: The resulting response object.
        """
        request = self._prepare_request(
            request=request, operation=operation, variables=variables, headers=headers
        )

        await self.validate(request=request)
        method = method or self._method

        if method == GraphQLQueryMethod.post:
            kwargs = dict(json=request.payload())
        elif method == GraphQLQueryMethod.get:
            kwargs = dict(params=request.payload(coerce=True))
        else:
            raise GraphQLClientException(f"Invalid method ({method}) specified")

        if session or self._session:
            return await self._http_request(
                session=session or self._session,
                method=method,
                request=request,
                **kwargs,
            )

        connector = await create_default_connector()
        async with aiohttp.ClientSession(connector=connector) as session:
            return await self._http_request(
                session=session, method=method, request=request, **kwargs
            )

    async def post(
        self,
        request: GraphQLRequest,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> GraphQLResponse:
        """
        Helper method that wraps `GraphQLClient.query` with method explicitly set as
        :attr:`GraphQLQueryMethod.post`.

        :param request: Request to send to the GraphQL server.
        :param headers: Additional headers to be set when sending HTTP request.
        :param operation: GraphQL operation name to use if the `GraphQLRequest.query`
                          contains named operations. This will override any default
                          operation set.
        :param variables: Query variables to set for the provided request. This will
                          override the default values for any existing variables in the
                          request if set.
        :param session: Optional `aiohttp.ClientSession` to use for requests
        :return: The resulting `GraphQLResponse` object.
        """
        return await self.query(
            request,
            method=GraphQLQueryMethod.post,
            headers=headers,
            operation=operation,
            variables=variables,
            session=session,
        )

    async def get(
        self,
        request: GraphQLRequest,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> GraphQLResponse:
        """
        Helper method that wraps :method: `GraphQLClient.query` with method explicitly
        set as :attr:`GraphQLQueryMethod.get`.

        :param request: Request to send to the GraphQL server.
        :param headers: Additional headers to be set when sending HTTP request.
        :param operation: GraphQL operation name to use if the `GraphQLRequest.query`
                          contains named operations. This will override any default
                          operation set.
        :param variables: Query variables to set for the provided request. This will
                          override the default values for any existing variables in the
                          request if set.
        :param session: Optional `aiohttp.ClientSession` to use for requests
        :return: The resulting `GraphQLResponse` object.
        """
        return await self.query(
            request,
            method=GraphQLQueryMethod.get,
            headers=headers,
            operation=operation,
            variables=variables,
            session=session,
        )

    async def subscribe(
        self,
        request: GraphQLRequest,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
        callbacks: Optional[CallbacksType] = None,
        on_data: Optional[CallbackType] = None,
        on_error: Optional[CallbackType] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> GraphQLSubscription:
        """
        Create and initialise a GraphQL subscription. Once subscribed and a known event
        is received, all registered callbacks for the event type is triggered with the
        :class:`aiographql.client.GraphQLSubscriptionEvent` instance passed in the first
        argument.

        The following example will start a subscription that prints all data events as
        it receives them.

        .. code-block:: python

            # initialise and subscribe to events in the background
            subscription: GraphQLSubscription = await client.subscribe(
                request="{ notifications: { id, summary } }",
                on_data=lambda event: print(f"Data: {event}"),
                on_error=lambda event: print(f"Error: {event}"),
            )
            # process events for 10 seconds then unsubscribe
            await asyncio.wait(subscription.task, timeout=10)
            subscription.unsubscribe()

        :param request: Request to send to the GraphQL server.
        :param headers: Additional headers to be set when sending HTTP request.
        :param operation: GraphQL operation name to use if the `GraphQLRequest.query`
                          contains named operations. This will override any default
                          operation set.
        :param variables: Query variables to set for the provided request. This will
                          override the default values for any existing variables in the
                          request if set.
        :param session: Optional `aiohttp.ClientSession` to use for requests
        :return: The resulting `GraphQLResponse` object.
        :param callbacks: Custom callback registry mapping an event to one more more
            callback methods. If not provided, a new instance is created.
        :param on_data: Callback to use when data event is received.
        :param on_error: Callback to use when an error occurs.
        :param session: Optional session to use for connecting the graphql endpoint, if
            one is not provided, a new session is created for the duration of the
            subscription.
        :return: The initialised subscription.
        """
        request = self._prepare_request(
            request=request, operation=operation, variables=variables, headers=headers
        )
        await self.validate(request=request)

        callbacks = callbacks or CallbackRegistry()
        if on_data:
            callbacks.register(GraphQLSubscriptionEventType.DATA, on_data)
        if on_error:
            callbacks.register(GraphQLSubscriptionEventType.ERROR, on_error)

        subscription = GraphQLSubscription(request=request, callbacks=callbacks)
        await subscription.subscribe(
            endpoint=self.endpoint, session=session or self._session
        )
        return subscription



from dataclasses import dataclass, field
from typing import Dict, Any, List

from aiographql.client.error import GraphQLError
from aiographql.client.request import GraphQLRequestContainer


@dataclass(frozen=True)
class GraphQLBaseResponse(GraphQLRequestContainer):
    json: Dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class GraphQLResponse(GraphQLBaseResponse):
    """
    GraphQL Response object wrapping response data and any errors. This object also
    contains the a copy of the :class:`GraphQLRequest` that produced this response.
    """

    @property
    def errors(self) -> List[GraphQLError]:
        """
        A list of :class:`GraphQLError` objects if server responded with query errors.
        """
        return [GraphQLError(**error) for error in self.json.get("errors", list())]

    @property
    def data(self) -> Dict[str, Any]:
        """The data payload the server responded with."""
        return self.json.get("data", dict())

    @property
    def query(self) -> str:
        """The query string used to produce this response."""
        return self.request.query



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

import graphql
import ujson as json


if TYPE_CHECKING:
    from aiographql.client import GraphQLResponse


class GraphQLClientException(Exception):
    pass


class GraphQLClientValidationException(GraphQLClientException):
    def __init__(self, *args: graphql.GraphQLError) -> None:
        message = "Query validation failed\n"
        for error in args:
            message += f"\n{str(error)}"
        super().__init__(message)


class GraphQLRequestException(GraphQLClientException):
    def __init__(self, response: GraphQLResponse) -> None:
        super().__init__(f"Request failed with response {json.dumps(response.json)}")
        self.response = response


class GraphQLIntrospectionException(GraphQLClientException):
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "Something went wrong during introspection process")



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field, asdict, replace, InitVar
from typing import Optional, Dict, Any

import ujson as json


@dataclass(frozen=True)
class GraphQLRequest:
    """
    GraphQL Request object that can be reused or used to store multiple named queries
    with default *operationName`, *variables* and *header* to use.

    :param query: GraphQL query string.
    :param operation: Optional name of operation to use from the query.
    :param variables: Variable dictionary pass with the query to the server.
    :param validate: If `True`, the request query is validated against the latest available
        schema from the server.
    :param headers: Headers to use, in addition to client default headers when making
        the HTTP request.
    """

    query: str
    operation: InitVar[Optional[str]] = field(default=None)
    operationName: Optional[str] = field(default=None, init=False)
    variables: Dict[str, Any] = field(default_factory=dict)
    validate: bool = field(default=True)
    headers: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self, operation: Optional[str] = None):
        if operation is not None:
            object.__setattr__(self, "operationName", operation)

    def __getattr__(self, item):
        if item == "operation":
            return self.operationName
        return super(GraphQLRequest, self).__getattribute__(item)

    @staticmethod
    def _coerce_value(value: Any) -> Any:
        if isinstance(value, bool):
            return int(value)
        if isinstance(value, dict):
            return json.dumps(value)
        return value

    def payload(self, coerce: bool = False) -> Dict[str, Any]:
        return {
            k: v if not coerce else self._coerce_value(v)
            for k, v in asdict(self).items()
            if v is not None and k not in {"schema", "headers", "validate"}
        }

    def copy(
        self,
        headers: Optional[Dict[str, str]] = None,
        headers_fallback: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
    ) -> GraphQLRequest:
        return replace(
            self,
            operation=operation or self.operationName,
            variables={**deepcopy(self.variables), **(variables or dict())},
            headers={
                **(headers_fallback or dict()),
                **self.headers,
                **(headers or dict()),
            },
        )


@dataclass(frozen=True)
class GraphQLRequestContainer:
    request: GraphQLRequest
    headers: InitVar[Optional[Dict[str, str]]] = field(default=None)
    operation: InitVar[Optional[str]] = field(default=None)
    variables: InitVar[Optional[Dict[str, Any]]] = field(default=None)

    def __post_init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
    ):
        object.__setattr__(
            self,
            "request",
            GraphQLRequest(query=self.request)
            if isinstance(self.request, str)
            else self.request.copy(
                headers=headers, operation=operation, variables=variables
            ),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiographql', 'aiographql.client']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.5,<4.0',
 'cafeteria-asyncio>=0.2,<0.3',
 'graphql-core>=3.0,<4.0',
 'ujson>=1.35,<2.0']

setup_kwargs = {
    'name': 'aiographql-client',
    'version': '1.0.0a0',
    'description': 'An asyncio GraphQL client built on top of aiohttp and graphql-core-next',
    'long_description': '# Asyncio GraphQL Client\n[![Documentation Status](https://readthedocs.org/projects/aiographql-client/badge/?version=latest)](https://aiographql-client.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/aiographql-client.svg)](https://badge.fury.io/py/aiographql-client)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Sonarcloud Status](https://sonarcloud.io/api/project_badges/measure?project=twyla-ai_aiographql-client&metric=alert_status)](https://sonarcloud.io/dashboard?id=twyla-ai_aiographql-client)\n[![](https://github.com/twyla-ai/aiographql-client/workflows/Test%20Suite/badge.svg)](https://github.com/twyla-ai/aiographql-client/actions?query=workflow%3A%22Test+Suite%22)\n\nAn asyncio GraphQL client built on top of aiohttp and graphql-core-next. The client by default introspects schemas and validates all queries prior to dispatching to the server.\n\n## Installation\n`pip install aiographql-client`\n\n## Example Usage\nHere are some example usages of this client implementation. These examples use the [Hasura GraphQL Engine](https://hasura.io/).\n\n### Simple Query\n```py\nasync def get_bots():\n    client = GraphQLClient(\n        endpoint="http://localhost:8080/v1alpha1/graphql",\n        headers={"x-hasura-admin-secret": "myadminsecretkey"},\n    )\n    request = GraphQLRequest(\n        query="""\n        query get_bots {\n          chatbot {\n            id, bot_name\n          }\n        }\n        """\n    )\n    transaction = await client.post(request)\n\n    # display the query used\n    print(transaction.request.query)\n\n    # dump the response data\n    print(transaction.response.data)\n```\n\n### Query Subscription\n```py\nasync def get_bots():\n    client = GraphQLClient(\n        endpoint="http://localhost:8080/v1alpha1/graphql",\n        headers={"x-hasura-admin-secret": "myadminsecretkey"},\n    )\n    request = GraphQLRequest(\n        query="""\n        subscription get_bot_updates {\n          chatbot {\n            id, bot_name\n          }\n        }\n        """\n    )\n    \n    # configure callbacks, here we simply print the event message when a data event\n    # (`GraphQLSubscriptionEvent`) is received.\n    callbacks = CallbackRegistry()\n    callbacks.register(\n        GraphQLSubscriptionEventType.DATA, lambda event: print(event.message)\n    )\n    \n    subscription: GraphQLSubscription = await client.subscribe(request, callbacks)\n    await subscription.task\n```\n\n### Query Validation Failures\nIf your query is invalid, thanks to graphql-core-next, we get a detailed exception in the traceback.\n\n```\naiographql.client.exceptions.GraphQLClientValidationException: Query validation failed\n\nCannot query field \'ids\' on type \'chatbot\'. Did you mean \'id\'?\n\nGraphQL request (4:13)\n3:           chatbot {\n4:             ids, bot_names\n               ^\n5:           }\n\nCannot query field \'bot_names\' on type \'chatbot\'. Did you mean \'bot_name\' or \'bot_language\'?\n\nGraphQL request (4:18)\n3:           chatbot {\n4:             ids, bot_names\n                    ^\n5:           }\n\n```\n\n### Sample Requests\n#### Using `variables` and `operationName`\n```py\nGraphQLRequest(\n    query="""\n    query get_bot_created($id: Int) {\n      chatbot(where: {id: {_eq: $id}}) {\n        id, created\n      }\n    }\n    query get_bot_name($id: Int) {\n      chatbot(where: {id: {_eq: $id}}) {\n        id, bot_name\n      }\n    }\n    """,\n    variables={\'id\': \'109\'},\n    operationName=\'get_bot_name\'\n)\n```\n',
    'author': 'Twyla Engineering',
    'author_email': 'software@twyla.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/twyla-ai/aiographql-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, NoReturn, Optional, Union

import aiohttp
from cafeteria.asyncio.callbacks import (
    CallbackRegistry,
    CallbackType,
    SimpleTriggerCallback,
)

from aiographql.client.helpers import create_default_connector
from aiographql.client.request import GraphQLRequestContainer
from aiographql.client.response import GraphQLBaseResponse, GraphQLResponse


class GraphQLSubscriptionEventType(Enum):
    """
    GraphQL Subscription Event Types
    """

    CONNECTION_INIT = "connection_init"
    CONNECTION_ACK = "connection_ack"
    CONNECTION_ERROR = "connection_error"
    CONNECTION_TERMINATE = "connection_terminate"
    START = "start"
    DATA = "data"
    ERROR = "error"
    COMPLETE = "complete"
    STOP = "stop"
    KEEP_ALIVE = "ka"


CallbacksType = Union[
    CallbackRegistry,
    Dict[GraphQLSubscriptionEventType, Union[CallbackType, List[CallbackType]]],
]


@dataclass(frozen=True)
class GraphQLSubscriptionEvent(GraphQLBaseResponse):
    """
    GraphQL subscription event wrapping the payload received from the server.

    :param subscription_id: The id of the subscription that generated this event.
    """

    subscription_id: Optional[str] = field(default=None)

    @property
    def id(self) -> Optional[str]:
        """The id of the event, if available."""
        return self.json.get("id")

    @property
    def type(self) -> Optional[GraphQLSubscriptionEventType]:
        """The type of event (:class:`GraphQLSubscriptionEventType`)."""
        try:
            return GraphQLSubscriptionEventType(self.json.get("type"))
        except ValueError:
            pass

    @property
    def payload(self) -> Optional[Union[GraphQLResponse, str]]:
        """The id of the subscription that generated this event."""
        payload = self.json.get("payload")
        if payload is not None:
            if self.type in (
                GraphQLSubscriptionEventType.DATA,
                GraphQLSubscriptionEventType.ERROR,
            ):
                return GraphQLResponse(request=self.request, json=payload)
            return payload


@dataclass(frozen=True)
class GraphQLSubscription(GraphQLRequestContainer):
    """
    Subscription container, with an attached
    :class:`cafeteria.asyncio.callbacks.CallbackRegistry`. When subscribed,
    the `task` will be populated with the :class:`asyncio.Task` instance.

    By default the subscription will be stopped, if an error, connection error or
    complete (:class:`GraphQLSubscriptionEventType`) is received.

    Subscription instances are intended to be used as immutable objects. However,
    `callbacks` and `stop_event_types` can be updated after initialisation.

    :param id: A unique subscription identifier that will be passed into any events
        generated by this subscription.
    :param callbacks: A :class:`CallbackRegistry` containing a mapping of
        :class:`GraphQLSubscriptionEventType` callback methods to trigger.
    :param stop_event_types: Events that cause the subscription to stop. By default,
        connection error, query error or connection complete events received are
        considered stop events.
    """

    id: str = field(default_factory=lambda: str(uuid.uuid4()), init=False)
    callbacks: Optional[CallbacksType] = field(default_factory=CallbackRegistry)
    stop_event_types: List[GraphQLSubscriptionEventType] = field(
        default_factory=lambda: [
            GraphQLSubscriptionEventType.ERROR,
            GraphQLSubscriptionEventType.CONNECTION_ERROR,
            GraphQLSubscriptionEventType.COMPLETE,
        ]
    )
    task: asyncio.Task = field(default=None, init=False, compare=False)

    def __post_init__(
        self,
        headers: Optional[Dict[str, str]] = None,
        operation: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
    ):
        super().__post_init__(headers, operation, variables)

        if self.callbacks is None:
            object.__setattr__(self, "callbacks", CallbackRegistry())
        elif isinstance(self.callbacks, dict):
            object.__setattr__(
                self, "callbacks", CallbackRegistry(callbacks=self.callbacks)
            )

    def active(self) -> bool:
        """
        Check if the subscription is active.

        :return: `True` if subscribed and active.
        """
        return (
            self.task is not None and not self.task.done() and not self.task.cancelled()
        )

    def connection_init_request(self) -> Dict[str, Any]:
        """
        Connection init payload to use when initiating a new subscription.

        :return: Connection initialise payload.
        """
        return {
            "type": GraphQLSubscriptionEventType.CONNECTION_INIT.value,
            "payload": {"headers": {**self.request.headers}},
        }

    def connection_start_request(self) -> Dict[str, Any]:
        """
        Connection start payload to use when starting a subscription.

        :return: Connection start payload.
        """
        return {
            "id": self.id,
            "type": GraphQLSubscriptionEventType.START.value,
            "payload": self.request.payload(),
        }

    def connection_stop_request(self) -> Dict[str, Any]:
        """
        Connection stop payload to use when stopping a subscription.

        :return: Connection stop payload.
        """
        return {"id": self.id, "type": GraphQLSubscriptionEventType.STOP.value}

    def is_stop_event(self, event: GraphQLSubscriptionEvent) -> bool:
        """
        Check if the provided *event* is configured as a stop even for this subscription.

        :param event: Event to check.
        :return: `True` if `event` is in `stop_event_types`.
        """
        return event.type in self.stop_event_types

    async def handle(self, event: GraphQLSubscriptionEvent) -> NoReturn:
        """
        Helper method to dispatch any configured callbacks for the specified event type.

        :param event: Event to dispatch callbacks for.
        """
        if event.id is None or event.id == self.id:
            self.callbacks.dispatch(event.type, event)

    async def _websocket_connect(
        self, endpoint: str, session: aiohttp.ClientSession
    ) -> None:
        """
        Helper method to create websocket connection with specified *endpoint*
        using the specified :class:`aiohttp.ClientSession`. Once connected, we
        initialise and start the GraphQL subscription; then wait for any incoming
        messages. Any message received via the websocket connection is cast into
        a :class:`GraphQLSubscriptionEvent` instance and dispatched for handling via
        :method:`handle`.

        :param endpoint: Endpoint to use when creating the websocket connection.
        :param session: Session to use when creating the websocket connection.
        """
        async with session.ws_connect(endpoint) as ws:
            await ws.send_json(data=self.connection_init_request())

            self.callbacks.register(
                GraphQLSubscriptionEventType.CONNECTION_ACK,
                SimpleTriggerCallback(
                    function=ws.send_json, data=self.connection_start_request()
                ),
            )

            try:
                async for msg in ws:  # type:  aiohttp.WSMessage
                    if msg.type != aiohttp.WSMsgType.TEXT:
                        if msg.type == aiohttp.WSMsgType.ERROR:
                            break
                        continue

                    event = GraphQLSubscriptionEvent(
                        subscription_id=self.id, request=self.request, json=msg.json(),
                    )
                    await self.handle(event=event)

                    if self.is_stop_event(event):
                        break
            except (asyncio.CancelledError, KeyboardInterrupt):
                await ws.send_json(data=self.connection_stop_request())

    async def _subscribe(
        self, endpoint: str, session: Optional[aiohttp.ClientSession] = None
    ) -> None:
        """
        Helper method wrapping :method:`GraphQLSubscription._websocket_connect` handling
        unique :class:`aiohttp.ClentSession` creation if on is not already provided.

        :param endpoint: Endpoint to use when creating the websocket connection.
        :param session: Optional session to use when creating the websocket connection.
        """
        if session:
            return await self._websocket_connect(endpoint=endpoint, session=session)

        connector = await create_default_connector()
        async with aiohttp.ClientSession(connector=connector) as session:
            return await self._websocket_connect(endpoint=endpoint, session=session)

    async def subscribe(
        self,
        endpoint: str,
        force: bool = False,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        """
        Create a websocket subscription and set internal task.

        :param endpoint: GraphQL endpoint to subscribe to
        :param force: Force re-subscription if already subscribed
        :param session: Optional session to use for requests
        """
        if self.active() and not force:
            return
        self.unsubscribe()
        task = asyncio.create_task(self._subscribe(endpoint=endpoint, session=session))
        object.__setattr__(self, "task", task)

    def unsubscribe(self) -> None:
        """
        Unsubscribe current websocket subscription if active and clear internal task.
        """
        if self.active():
            try:
                self.task.cancel()
            except asyncio.CancelledError:
                pass
        object.__setattr__(self, "task", None)

    async def unsubscribe_and_wait(self) -> None:
        task = self.task
        self.unsubscribe()
        try:
            await task
        except asyncio.CancelledError:
            pass
